// Draw data using a star glyph pattern.  Each dimension of a data point
// controls the length of a ray emanating from a point, and the ends of all
// N rays are connected with a polyline.  Glyphs are currently spaced
// evenly across the drawing area, using the square root of the data set
// size to determine the number of rows and columns of glyphs.  We could also
// use 2 dimensions to control the location of the glyphs, but you tend to
// get a lot of overlap.

#include <xmdv/CFlatGlyph.h>
#include <xmdv/ogl.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/brush.h>
#include <xmdv/brush_oper.h>
#include <xmdv/multi_select.h>
#include <string.h>
#include <xmdv/glyph_place.h>
#include <stdlib.h>

CFlatGlyph::CFlatGlyph(CFlatDisplayInformation* flat_info) : CFlatDisplay(flat_info)
{
}

CFlatGlyph::~CFlatGlyph()
{
}


//
// Redraws a single glyph on the canvas.  Usefull when
// doing multi-select for example.
//
void CFlatGlyph::RedrawSingleGlyph(grCanvas *canvas, int glyphno)
{
    ARR<double> data(dataset->dims);
    dataset->GetData(data, glyphno);

    DrawSingleData(data,glyphno);
}


//set color and style for unbrushed data
void CFlatGlyph::SetUnbrushedColorAndStyle()
{
    CFlatDisplay:: SetUnbrushedColorAndStyle();
    glLineWidth((float) DSIZE);
}



void CFlatGlyph::InputGlyphBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (curr_brush->display == false || curr_brush->enable == false)
        return;

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    double halfbsize = 0.5/(1.0 + sqrt(dataset->data_size));

	int i, glyph_no = -1;
	double min = halfbsize *3, temp;
	for ( i = 0; i < dataset->data_size; i++ )
	{
		if ( (fabs(p.X - dataset->basex[i]) < halfbsize) && 
			 (fabs(p.Y - dataset->basey[i]) < halfbsize) ) 
		{
			temp = fabs(p.X - dataset->basex[i]) + fabs(p.Y - dataset->basey[i]);
			if ( temp < min )
			{
				glyph_no = i; 
				min = temp;
			}
		}
	}

	if ( glyph_no == -1 )
		return;

    if (multi_select)
    {
        // Delete point if already there
        if (multiSelectDeletePoint(glyph_no)==true)
        {
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[DATA]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        } else
        {
            // add this point
            multiSelectAddPoint(glyph_no);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        }
        return;
    }

    ARR<double> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    for (i=0; i<dataset->dims; i++)
        curr_brush->pos[i] = data[i];

    // Check if any of the brush values went out of range
    for (i=0; i<dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (curr_brush->Max(i) > dataset->dim_max[i])
            curr_brush->pos[i] = dataset->dim_max[i] - curr_brush->size[i]/2.;
        if (curr_brush->Min(i) < dataset->dim_min[i])
            curr_brush->pos[i] = dataset->dim_min[i] + curr_brush->size[i]/2.;
    }
}

void CFlatGlyph::InputGlyphButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    if (button==2) {
        canvas->SetHandCursor();
        return;
    }

    if (button == 0) {
        InputGlyphBrushSelect(canvas,x,y);
        BrushOperMarkDataset();
    }

    if (!multi_select)
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
}

//jing need to do for distorton?
static void GlyphCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    double halfbsize = 0.5/(1.0 + sqrt(dataset->data_size));

	int i, glyph_no = -1;
	double min = halfbsize *3, temp;
	for ( i = 0; i < dataset->data_size; i++ )
	{
		if ( (fabs(p.X - dataset->basex[i]) < halfbsize) && 
			 (fabs(p.Y - dataset->basey[i]) < halfbsize) ) 
		{
			temp = fabs(p.X - dataset->basex[i]) + fabs(p.Y - dataset->basey[i]);
			if ( temp < min )
			{
				glyph_no = i; 
				min = temp;
			}
		}
	}

	if ( glyph_no == -1 )
		return;

    ARR<double> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    char caption[50000]="\0";
    sprintf(caption, "Glyph %d: (", glyph_no);

    for (i=0; i<dataset->dims; i++) {
        char str[100];

        if (dataset->nominal_fields[i] != 0) {
//            int value = (int) rint(data[i]);
//jing
            int value = (int)(data[i]+0.5);
//jing
            const char *string = dataset->nominal_fields[i]->MapValueToString(value);
            sprintf(str, "%s=%s ", dataset->names[i], string);
        } else {
            sprintf(str, "%s=%.2lf ", dataset->names[i], data[i]);
        }

        strcat(caption, str);
    }
    strcat(caption,")");

 //   appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),caption, canvas->name);
}



void CFlatGlyph::InputGlyphMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    GlyphCaption(canvas, x, y);

    if (canvas->button[2] == true) {
        appCanvasMousePan(canvas);
    }
}

void InputGlyphButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (canvas->shiftkey) {
        return;
    }

    canvas->SetDefaultCursor();
}

void CFlatGlyph::SetParameters()
{
    CFlatDisplay::SetParameters();
    
    norm.SetLen(dataset->dims);

    int poly_length;
//    if ( !show_two_axes )
        poly_length = dataset->dims + 1;
/*    else
    {
        poly_length = 1;
        for ( int i = 0; i < dataset->dims; i++ )
        {
            if ( dim_reducer->IsLeafDim(i) )
                poly_length ++;
            else 
                poly_length += 3;
        }
    } */
    poly.SetLen(poly_length);
}

//Add zhuo's code here
void CFlatGlyph::DrawSingleData(ARR<double> data, int datano)
{
    double blocksize = dataset->blocksize;
    double basex = dataset->basex[datano];
    double basey = dataset->basey[datano];
    
    //jing
    /*RGBt c(interface_color[DATA]);
    glColor3f((float) c.R, (float) c.G, (float) c.B );
    */

    // Compute normalized length of each ray for sample
    //
    int j;
    double last_x, last_y;     // ZFC

    for (j=0; j<dataset->dims; j++) {
        norm[j] = (data[j] - dataset->dim_min[j]) /
                (dataset->dim_max[j] - dataset->dim_min[j]);
    }

    // Compute endpoints for each ray based on evenly spaced angles
    // from center
    //
    double alpha = 2.0*M_PI/dataset->dims;
    double center = ((dataset->dims-1)*alpha/70)/2;  // ZFC for centering the Profile

    //Draw the Profile Shape
    if (flat_info->glyphShape == 1) { 
        last_x = basex;
        
        last_y = basey;
        for (j=0; j<dataset->dims; j++) {
            poly[j].X = basex + (j * alpha/70) ;
            poly[j].Y = basey - (norm[j] * blocksize);
            canvas->DrawLine(last_x-center, last_y+center, poly[j].X-center, poly[j].Y+center);
            last_x = poly[j].X;
            last_y = poly[j].Y;
        }
    
        glLineWidth(1.);
        // Draw bounding polyline
        canvas->DrawLine(last_x-center, basey+center, last_x-center, last_y+center);
        canvas->DrawLine(last_x-center, basey+center, basex-center,  basey+center);

    //Draw the Star Shape
    } else {
        for (j=0; j<dataset->dims; j++) 
        {
            if ( show_wide_axes )
            {
                if ( dim_reducer->IsLeafDim(j) )
                    glLineWidth(1.);
                else 
                {
                    double dim_disagreement = dim_reducer->GetDimDisagreement(j);
                    glLineWidth((float)(1.+ dim_disagreement));
                }
            }
            else
                glLineWidth(1.);

            poly[j].X = basex + norm[j] * screen_distortion->star_angles_cos_sized[j];
            poly[j].Y = basey - norm[j] * screen_distortion->star_angles_sin_sized[j];
            canvas->DrawLine(basex, basey, poly[j].X, poly[j].Y);
        }

        // Draw bounding polyline
        poly[dataset->dims].X = poly[0].X;
        poly[dataset->dims].Y = poly[0].Y;

        canvas->DrawLines(poly);
    }
}

void CFlatGlyph::DrawSingleDataWithTwoAxes(ARR<double> mean, ARR<double> max, 
                                           ARR<double> min, int datano, bool unbrushed)
{
    //jing
    /*RGBt c(interface_color[DATA]);
    glColor3f((float) c.R, (float) c.G, (float) c.B );
    */

    double blocksize = dataset->blocksize;
    double basex = dataset->basex[datano];
    double basey = dataset->basey[datano];

    // Compute normalized length of each ray for sample
    //
    int i, j;

    for (j=0; j<dataset->dims; j++) {
        norm[j] = (data[j] - dataset->dim_min[j]) /
                (dataset->dim_max[j] - dataset->dim_min[j]);
    }

    // Compute endpoints for each ray based on evenly spaced angles
    // from center
    //
    double alpha = 2.0*M_PI/dataset->dims;
    double center = ((dataset->dims-1)*alpha/70)/2;  // ZFC for centering the Profile

    //Draw the Profile Shape
    if (flat_info->glyphShape == 1) { 
/*        last_x = basex;
        
        last_y = basey;
        for (j=0; j<dataset->dims; j++) {
            poly[j].X = basex + (j * alpha/70) ;
            poly[j].Y = basey - (norm[j] * 2);
            canvas->DrawLine(last_x-center, last_y+center, poly[j].X-center, poly[j].Y+center);
            last_x = poly[j].X;
            last_y = poly[j].Y;
        }
    
        // Draw bounding polyline
        canvas->DrawLine(last_x-center, basey+center, last_x-center, last_y+center);
        canvas->DrawLine(last_x-center, basey+center, basex-center,  basey+center);
*/
    } 
    //Draw the Star Shape
    else 
    {
        double normalized_mean;
        
        for (i=0; i<dataset->dims; i++) 
        {
            if ( dim_reducer->IsLeafDim(i) )
            {
                canvas->SetForeground(interface_color[DATA]);
                normalized_mean = (mean[i] - dataset->dim_min[i]) / (dataset->dim_max[i] - dataset->dim_min[i]);
                poly[i].X = basex + normalized_mean * screen_distortion->star_angles_cos_sized[i];
                poly[i].Y = basey - normalized_mean * screen_distortion->star_angles_sin_sized[i];
                canvas->DrawLine(basex, basey, poly[i].X, poly[i].Y);
            }
            else
            {
                //mean, min and max of un-leaf nodes neednot be normalized, since they are from 0-1.
                canvas->SetForeground(interface_color[DATA]);
                poly[i].X = basex + mean[i] * screen_distortion->star_angles_cos_sized[i];
                poly[i].Y = basey - mean[i] * screen_distortion->star_angles_sin_sized[i];
                canvas->DrawLine(basex, basey, poly[i].X, poly[i].Y);

                Vec2 p;
                canvas->SetForeground(interface_color[GRID1]);
                p.X = basex + min[i] * screen_distortion->star_angles_cos_sized[i];
                p.Y = basey - min[i] * screen_distortion->star_angles_sin_sized[i];
                canvas->DrawLine(p.X, p.Y, poly[i].X, poly[i].Y);

                canvas->SetForeground(interface_color[GRID2]);
                p.X = basex + max[i] * screen_distortion->star_angles_cos_sized[i];
                p.Y = basey - max[i] * screen_distortion->star_angles_sin_sized[i];
                canvas->DrawLine(p.X, p.Y, poly[i].X, poly[i].Y);
            }
        }

        // Draw bounding polyline
        if ( unbrushed )
            canvas->SetForeground(interface_color[DATA]);
        else
            canvas->SetForeground(dataset->color[datano]);
        poly[i].X = poly[0].X;
        poly[i].Y = poly[0].Y;

        canvas->DrawLines(poly);
    }
}

void CFlatGlyph::DrawOthers()
{
    if ( ! screen_distortion->link )
        return;

    double* prin0;
    dataset->GetPrin0(prin0);

    int data_size = dataset->data_size;

    SortStruct *sortArray = new SortStruct[data_size];

    int i;
//    ARR<double> data;
//    data.SetLen(dims);

    for ( i = 0; i < data_size; i++ )			
	{
//        dataset->GetData(data, i);
		sortArray[i].number = i;
//		sortArray[i].value = data[sort_dim];
		sortArray[i].value = prin0[i];
	}

    qsort( sortArray, data_size, sizeof( SortStruct ), IncreaseSort );

//    int sort_dim = screen_distortion->sort_dim;
//    double thres_value = (dataset->dim_max[sort_dim] - dataset->dim_min[sort_dim]) * screen_distortion->threshold; 
    double max, min;
    max = sortArray[data_size-1].value;
    min = sortArray[0].value;

    double thres_value = (max - min) * screen_distortion->threshold; 

    if ( screen_distortion->link_method == 1 )
    {
        int j, k, m;
        bool stop;
        for ( k = 0; k < data_size-1; k++ )
	    {
            i = sortArray[k].number;
            if (dataset->mark[i] == okcdata::DISABLED) continue;
/*        
            if (dataset->mark[i] == okcdata::REGULAR) 
                canvas->SetForeground(interface_color[HIGHLIGHT2]);
            else
                canvas->SetForeground(interface_color[HIGHLIGHT3]);
*/        
            stop = false;
            double first, second;
        
            dataset->GetData(data, i);
            first = prin0[i];
            m = k+1;
        
            while ( ! stop )
            {
                j = sortArray[m].number;
                if (dataset->mark[j] != okcdata::DISABLED)
                {
//                    dataset->GetData(data, j);
                    second = prin0[j];
                    if ( (second - first) <= thres_value )
                    {
                        if (dataset->mark[j] != okcdata::REGULAR || dataset->mark[i] != okcdata::REGULAR) 
                        {
                            canvas->SetForeground(interface_color[HIGHLIGHT3]);
                            canvas->DrawLine(dataset->basex[i], dataset->basey[i], dataset->basex[j], dataset->basey[j]); 
                            if ( dataset->mark[j] == okcdata::REGULAR && dataset->mark[i] != okcdata::REGULAR)
                            {
                                canvas->SetForeground(interface_color[HIGHLIGHT4]);
                                RedrawSingleGlyph(canvas, j);
                            }
                            if ( dataset->mark[j] != okcdata::REGULAR && dataset->mark[i] == okcdata::REGULAR)
                            {
                                canvas->SetForeground(interface_color[HIGHLIGHT4]);
                                RedrawSingleGlyph(canvas, i);
                            }
                            if ( dataset->mark[i] != okcdata::REGULAR )
                            {
                                canvas->SetForeground(interface_color[HIGHLIGHT1]);
                                RedrawSingleGlyph(canvas, i);
                            }
                            if ( dataset->mark[j] != okcdata::REGULAR )
                            {
                                canvas->SetForeground(interface_color[HIGHLIGHT1]);
                                RedrawSingleGlyph(canvas, j);
                            }
                        }
                         //   else
                           // canvas->SetForeground(interface_color[HIGHLIGHT2]);

                    }
                    else 
                    {
                        stop = true;
                    }
                }

                m++;
                if ( m >= data_size )
                {
                    stop = true;
                }
            }
	    }
    }
    else
    {
        int j, k;
        for ( k = 0; k < data_size-1; k++ )
        {
            i = sortArray[k].number;
            j = sortArray[k+1].number;
            canvas->DrawLine(dataset->basex[i], dataset->basey[i], dataset->basex[j], dataset->basey[j]); 
        }
    }

	delete sortArray;
    delete prin0;

}


void CFlatGlyph::InputMouseMotion(int x, int y)
{
    InputGlyphMouseMotion(canvas,x,y);
}


void CFlatGlyph::InputButtonPress(int button, int x, int y)
{
    InputGlyphButtonPress(canvas, button, x,y);
}

void CFlatGlyph::InputButtonRelease(int button, int x, int y)
{
    InputGlyphButtonRelease(canvas, button, x, y);
}


